// $Id: FittedRegion3D.cc,v 1.3 2005/04/11 12:47:38 heller Exp $
/* class FittedRegion3D
 *
 * This class adds boundary information as well as some useful output
 * methods to FittedPlane3D.
 *
 * created  Mon Nov 20 14:41:06 1995  Justus Piater
 */ 
#include "FittedRegion3D.h"

void FittedRegion3D::trace ()
{
  Register (inliers);
  Register (outliers);
  Register (hull);
  Register (holes);
}

void FittedRegion3D::init ()
{
  hull  = NULL;
  holes = NULL;
}

FittedRegion3D::FittedRegion3D () : FittedPlane3D ()
{
  init ();
}

FittedRegion3D::FittedRegion3D (FittedPlane3D &plane) : FittedPlane3D (plane)
{
  init ();
}

FittedRegion3D::FittedRegion3D (FittedRegion3D &)
{
  cerr << "No copy constructor for FittedRegion3D defined\n";
  exit (-1);
}

FittedRegion3D FittedRegion3D::operator= (FittedRegion3D &)
{
  cerr << "No assignment operator for FittedRegion3D defined\n";
  exit (-1);
}

FittedRegion3D::~FittedRegion3D ()
{
  delete hull;
  delete holes;
}

void FittedRegion3D::output(Isrostream& os)
{
  os << name () << " with coefficients (" 
     << "a=" << a << ",b=" << b << ",c=" << c << ",d=" << d
     << "),\nfit_error=" << fit_error
     << ",\ninliers=" << inliers
     << ",\noutliers=" << outliers
     << ",\nhull=" << hull
     << ",\nholes=" << holes << "\n";
}

void FittedRegion3D::input(Isristream& is)
{
  is >> name () >> " with coefficients (" 
     >> "a=" >> a >> ",b=" >> b >> ",c=" >> c >> ",d=" >> d
     >> "),"
     >> "fit_error=" >> fit_error >> ","
     >> "inliers=";
  inliers = (Point3DSet *) read_token_ptr(is);
  is >> "," >> "outliers=";
  outliers = (Point3DSet *) read_token_ptr(is);
  test_coeff();
  is >> "," >> "hull=";
  hull = (TokenList *) read_token_ptr (is);
  is >> "," >> "holes=";
  holes = (TokenSet *) read_token_ptr (is);
}

const int dwidth    = 12;
const int dprec     =  4;
const int n_per_row = 10;

#ifdef INVENTOR
void FittedRegion3D::output_inventor_normal (ofstream &os)
{
  // output normal
  double a_up, b_up, c_up;
  if (c < 0) {
    a_up = -a;
    b_up = -b;
    c_up = -c;
  } else {
    a_up = a;
    b_up = b;
    c_up = c;
  }
  os << "  Normal { vector [ " << a_up << " " << b_up << " " << c_up
     << " ] }\n  NormalBinding { value PER_FACE }\n";

  if (left_handed) os << "  Transform { scaleFactor 1 -1 1 }\n";
}


void FittedRegion3D::output_inventor (ofstream &os, bool points, bool poles)
/* outputs inventor code, eg. for SceneViewer.
   If points==true, the points are output as well (inliers green,
   outliers red).
   If poles==true, a line is drawn from each point to its projected
   location on the plane. */
{
  if (!points && !poles) {
    output_inventor (os);
    return;
  }

  Point3D *point;
  int inPoints   = inliers  ? inliers->count ()  : 0;
  int outPoints  = outliers ? outliers->count () : 0;
  int i, nPoints = inPoints + outPoints;
  TokenArray *array = new TokenArray (NULL, nPoints);
  if (inliers ) array->Union (inliers );
  if (outliers) array->Union (outliers);

  os << "#Inventor V2.0 ascii\n\nSeparator {\n  Coordinate3 {\n    point [ "
     << setiosflags (ios::fixed) << setprecision (dprec);

  // output point coordinates
  for (i = 0; i < nPoints; i++) {
    if (i) os << "            ";
    point = (Point3D *) (*array)[i];
    os << setw (dwidth) << point->x << " "
       << setw (dwidth) << point->y << " "
       << setw (dwidth) << point->z << ",\n";
  }
  // output projected point coordinates
  double dist;
  for (i = 0; i < nPoints; i++) {
    point = (Point3D *) (*array)[i];
    dist = a*point->x + b*point->y + c*point->z + d;
    os << "            "
       << setw (dwidth) << point->x - dist*a << " "
       << setw (dwidth) << point->y - dist*b << " "
       << setw (dwidth) << point->z - dist*c;
    if (i < nPoints - 1) os << ",\n";
  }
  os << " ]\n  }\n";

  output_inventor_normal (os);

  // output plane indices
  os << "  IndexedFaceSet {\n    coordIndex [ ";
  TokenSetState *state = hull->state ();
  Token *hullpoint;
  i = -1;
  for (hullpoint = state->value (); hullpoint; hullpoint = state->next ()) {
    if (++i == n_per_row) {
      os << "\n                 ";
      i = 0;
    }
    os << array->index (hullpoint) + nPoints << ", ";
  }
  delete state;
  os << "-1 ]\n  }\n";

  if (poles) {			// output pole indices
    for (i = 0; i < nPoints; i++) {
      os << "  IndexedLineSet { coordIndex [ "
	 << i << ", " << i + nPoints << ", -1 ] }\n";
    }
  }

  if (points) {
    if (inliers) {
      os << "  Material {\n"
	 << "    ambientColor 0 1 0\n    diffuseColor 0 1 0\n  }\n"
	 << "  PointSet {\n    numPoints " << inPoints << "\n  }\n";
    }
    if (outliers) {
      os << "  Material {\n"
	 << "    ambientColor 1 0 0\n    diffuseColor 1 0 0\n  }\n"
	 << "  PointSet {\n    startIndex " << inPoints
	 << "\n    numPoints " << outPoints << "\n  }\n";
    }
  }

  // output rest
  os << "}\n";
  delete array;
}

void FittedRegion3D::output_inventor (ofstream &os)
/* This function is like the above, except that it saves lots of
   disk space (and computation time) by not outputting any
   point coordinates. */
{
  int i, n, nPoints = hull->count ();

  os << "#Inventor V2.0 ascii\n\nSeparator {\n  Coordinate3 {\n    point [ "
     << setiosflags (ios::fixed) << setprecision (dprec);

  // output projected point coordinates
  TokenSetState *state = hull->state ();
  Point3D *point;
  double dist;
  for (point = (Point3D *) state->value (), i = 0;
       point;
       point = (Point3D *) state->next (), i++) {
    dist = a*point->x + b*point->y + c*point->z + d;
    if (i) os << "            ";
    os << setw (dwidth) << point->x - dist*a << " "
       << setw (dwidth) << point->y - dist*b << " "
       << setw (dwidth) << point->z - dist*c;
    if (i < nPoints - 1) os << ",\n";
  }
  delete state;
  os << " ]\n  }\n";

  output_inventor_normal (os);

  // output plane indices
  os << "  IndexedFaceSet {\n    coordIndex [ ";
  i = -1;
  for (n = 0; n < nPoints; n++) {
    if (++i == n_per_row) {
      os << "\n                 ";
      i = 0;
    }
    os << setw (4) << n << ", ";
  }
  os << "  -1 ]\n  }\n";

  os << "}\n";
}

void FittedRegion3D::output_polygon (ofstream &os_xyz, ofstream &os_polygons,
				     bool raw)
/* writes to two files: one contains vertex coordinates, the other integers
   specifying the number of consecutive vertices belonging to this region.
   This output can be read by eg. IDL.
   Both ascii and binary (raw==true) output is available. */
{
  TokenSetState *state = hull->state ();
  Point3D *point;
  short svalue;			// 16-bit integer
  float fvalue;			// 32-bit floating points

  if (raw) {
    svalue = short (hull->count ());
    os_polygons.write (&svalue, sizeof (svalue));
  } else {
    os_polygons << setw (4) << hull->count () << endl;
    os_xyz << setiosflags (ios::fixed) << setprecision (dprec);
  }

  for (point = (Point3D *) state->value ();
       point;
       point = (Point3D *) state->next  ()) {
    if (raw) {
      fvalue = float (point->x);
      os_xyz.write (&fvalue, sizeof (fvalue));
      fvalue = float (point->y);
      os_xyz.write (&fvalue, sizeof (fvalue));
      fvalue = float (point->z);
      os_xyz.write (&fvalue, sizeof (fvalue));
    } else {
      os_xyz << setw (dwidth) << point->x << " "
	     << setw (dwidth) << point->y << " "
	     << setw (dwidth) << point->z << endl;
    }
  }
  delete state;
}

#endif
