#include <Plane3D.h>

/* Plane3D.cc
   Created 10/23/95 by James Doyle
   This module provides the Plane3D class and its derived FittedPlane3D
class, which represent planes in 3D space, along with optional
information about the results of fitting algorithms generating them.
*/

bool Plane3D::test_coeff()
{
  // make sure a,b,c give unit vector
  return validity = (Vector3D(a,b,c)).is_unit();
}

void Plane3D::d_check()
{
  // make sure eq has positive d
  if (d < 0)
    {
      a *= -1;
      b *= -1;
      c *= -1;
      d *= -1;
    }
}

Plane3D::Plane3D() : a(0), b(0), c(0), d(0), validity(false) {}

Plane3D::Plane3D(float newa, float newb, float newc, float newd)
: a(newa), b(newb), c(newc), d(newd)
{
  test_coeff();
  d_check();
}

Plane3D::Plane3D(Point3DArray three_points)
// generates plane coefficients from the given points
{

  // check that correct number of distinct points are there
  if (three_points.count() != 3) {
    validity = false;
    return;
  }
  Point3D p1 = (*(Point3D*)(three_points[0]));
  Point3D p2 = (*(Point3D*)(three_points[1]));
  Point3D p3 = (*(Point3D*)(three_points[2]));
  if ((p1 == p2)
      || (p2 == p3)
      || (p3 == p1)) {
    validity = false;
    return;
  }

  // make vectors for calculations
  Vector3D v1 = (Vector3D)p1,
           v2 = (Vector3D)p2,
           v3 = (Vector3D)p3;
  
  // generate vector and check validity
  Vector3D Normal = (v2 - v1).cross(v3 - v1);
  double Normals_magnitude = sqrt(Normal.dot(Normal));
  if (!(bool)Normals_magnitude) {
    validity = false;
    return;
  }

  // take coefficients from the unit vector corresponding to normal
  Vector3D Unit_norm = Normal / Normals_magnitude;
  a = Unit_norm.i;
  b = Unit_norm.j;
  c = Unit_norm.k;
  d = -(Unit_norm.dot(v1));

  validity = true;
  d_check();
}

ostream& operator<<(ostream& os, const Plane3D& pl)
{
  os << "Plane3D with coefficients (" 
    << "a=" << pl.a << ",b=" << pl.b << ",c=" << pl.c << ",d=" << pl.d
      << ")\n";
  return os;
}

void Plane3D::output(Isrostream& os)
{
  os << "Plane3D with coefficients (" 
    << "a=" << a << ",b=" << b << ",c=" << c << ",d=" << d
      << ")\n";
}

void Plane3D::input(Isristream& is)
{
  is >> "Plane3D with coefficients (" 
    >> "a=" >> a >> ",b=" >> b >> ",c=" >> c >> ",d=" >> d
      >> ")";
  test_coeff();
}

float Plane3D::perp_distance_from(const Point3D& pt) const
{
  // determine distance by plugging point into equation
  return fabs((a*pt.x) + (b*pt.y) + (c*pt.z) + d);
} 

float Plane3D::z_at(double x, double y) const
{
  // determine z by plugging x and y into equation
  return ((-(a*x) - (b*y) - d) / c); 
}

Vector3D Plane3D::normal() const
{
  return Vector3D(a,b,c);
}

FittedPlane3D::FittedPlane3D() 
: Plane3D(), fit_error(0), inliers(NULL), outliers(NULL) {}

FittedPlane3D::FittedPlane3D(float newa, float newb, float newc, float newd)
: Plane3D(newa, newb, newc, newd), fit_error(0), inliers(NULL), outliers(NULL) 
{}

FittedPlane3D::FittedPlane3D(Point3DArray three_points) 
: Plane3D(three_points), fit_error(0), inliers(NULL), outliers(NULL) {}

ostream& operator<<(ostream& os, const FittedPlane3D& pl)
{
  os << "FittedPlane3D with coefficients (" 
    << "a=" << pl.a << ",b=" << pl.b << ",c=" << pl.c << ",d=" << pl.d
      << "),\n"
	<< "fit_error=" << pl.fit_error << ",\n"
	  << "inliers=" << pl.inliers << ",\n"
	    << "outliers=" << pl.outliers << "\n";
  
  return os;
}

void FittedPlane3D::trace()
{
  Register(inliers);
  Register(outliers);
}

void FittedPlane3D::output(Isrostream& os)
{
  os << "FittedPlane3D with coefficients (" 
    << "a=" << a << ",b=" << b << ",c=" << c << ",d=" << d
      << "),\n"
	<< "fit_error=" << fit_error << ",\n"
	  << "inliers=" << inliers << "\n"
	    << "outliers=" << outliers << "\n";
}

void FittedPlane3D::input(Isristream& is)
{
  is >> "FittedPlane3D with coefficients (" 
    >> "a=" >> a >> ",b=" >> b >> ",c=" >> c >> ",d=" >> d
      >> "),"
	>> "fit_error=" >> fit_error >> ","
	  >> "inliers=";
  inliers = (Point3DSet*)read_token_ptr(is);
  is >> "outliers=";
  outliers = (Point3DSet*)read_token_ptr(is);
  test_coeff();
}

#ifdef OUTPUT_RESIDUALS
void FittedPlane3D::output_residuals (ofstream &os, bool binary)
{
  if (!binary) os << setiosflags (ios::fixed) << setprecision (4);

  TokenSetState *state = inliers->state ();
  Point3D *point;
  float residual;
  for (point = (Point3D *) state->value ();
       point;
       point = (Point3D *) state->next ()) {
    residual = (a * point->x +
		b * point->y +
		c * point->z + d);
    if (binary) os.write ((const unsigned char *)&residual, sizeof (residual));
    else os << setw (12) << residual<< endl;
  }
  delete state;
}
#endif

double FittedPlane3D::mean_squared_residual ()
{
  double residual, sum = 0.0;
  int count = 0;
  TokenSetState *state = inliers->state ();
  Point3D *point;

  for (point = (Point3D *) state->value ();
       point;
       point = (Point3D *) state->next ()) {
    residual = perp_distance_from (*point);
    sum += residual * residual;
    count++;
  }
  delete state;
  return sum / count;
}

static void add_coords (TokenSet *points, Point3D &sum_point)
{
  if (!points) return;

  TokenSetState *state = points->state ();
  Point3D *point;
  for (point = (Point3D *) state->value ();
       point;
       point = (Point3D *) state->next ()) {
    sum_point.x += point->x;
    sum_point.y += point->y;
    sum_point.z += point->z;
  }
  delete state;
}

Point3D *FittedPlane3D::find_center ()
{
  // compute center of mass of all points...
  Point3D *center = new Point3D ();
  add_coords (inliers, *center);
  add_coords (outliers, *center);
  int sum = inliers->count() + outliers->count();
  center->x /= sum;
  center->y /= sum;
  center->z /= sum;

  // ...and project it onto the fitted plane (perpendicularly)
  double l = a*center->x + b*center->y + c*center->z + d;
  center->x -= l * a;
  center->y -= l * b;
  center->z -= l * c;
  return center;
}

