// -*- C++ -*-
#ifndef PLANE3D_DEFINED
#define PLANE3D_DEFINED
#include <Token.h>
#include <Container.h>
#include <Point3D.h>
#include <Vector3D.h>

/* Plane3D.h
   Created 10/23/95 by James Doyle
   This module provides the Plane3D class and its derived FittedPlane3D
class.  Plane3D is the representation of a plane in 3D space using
coefficients a,b,c,d.  It can be constructed directly from coefficients
or from three non-collinear points.  It provides input and output and
methods to: find the perpendicular distance from a 3D point, find the
z value of the plane at an x,y pair, and return a unit vector normal
to the plane.  
   FittedPlane3D adds a fit_error field and two token arrays of points
that represent inliers and outliers.  In general, FittedPlane3D is meant
to contain data returned by a plane fitting module.  The fit error
represents some median or mean error of both inlier and outlier points
from the plane.  The inliers list represents points found to represent
the plane by the fitting algorithm, while the outliers list represents
other points in the sample set.  Input and output (including the 
point lists) are provided for FittedPlane3D.
*/

typedef TokenArray Point3DArray; 
typedef TokenSet Point3DSet;

class Plane3D : public Token {
  friend ostream& operator<<(ostream& os, const Plane3D& pl);

 public:
  float a, b, c, d;
  bool validity;

  Plane3D();
  Plane3D(float newa, float newb, float newc, float newd);
  Plane3D(Point3DArray three_points);

  char *name() const { return "Plane3D"; }

  void output(Isrostream& os);
  void input(Isristream& is);

  float perp_distance_from(const Point3D& pt) const;
  float z_at(double x, double y) const;
  Vector3D normal() const;

 protected:
  bool test_coeff();
  void d_check();
};

class FittedPlane3D : public Plane3D {
  friend ostream& operator<<(ostream& os, const FittedPlane3D& pl);
 public:
  float fit_error;
  Point3DSet* inliers;
  Point3DSet* outliers;

  FittedPlane3D();
  FittedPlane3D(float newa, float newb, float newc, float newd);
  FittedPlane3D(Point3DArray three_points);

  char *name() const { return "FittedPlane3D"; }

  void trace();
  void output(Isrostream& os);
  void input(Isristream& is);

  Point3D *find_center ();
  double mean_squared_residual ();
#ifdef OUTPUT_RESIDUALS
  void output_residuals (ofstream &os, bool binary);
#endif
};


#endif



